/*
 *	main.c
 *
 *	ColecoVision/EMU
 *
 *	* Sun Jan 23 00:00:00 JST 2005 Naoyuki Sawa
 *	- 1st [XB
 */
#include "app.h"

//#define TEST
#ifdef TEST
void test();
#endif /*TEST*/

/****************************************************************************
 *	
 ****************************************************************************/

#define extern
#include "work.h"
#undef extern

/****************************************************************************
 *	main
 ****************************************************************************/

int
main()
{
	/* B */
	app_init();

	/* C[vB */
#ifndef TEST
	app_main();
#else /*TEST*/
	test();
#endif /*TEST*/

	return 0;
}

/****************************************************************************
 *	app_init
 ****************************************************************************/

void
app_init()
{
	/* ʓIȏB */
	pceLCDDispStop();
	pceLCDSetBuffer(vbuff);
	pceAppSetProcPeriod(1000 / FPS);
	pcePadSetTrigMode(PP_MODE_SINGLE);
	pceLCDDispStart();

	/* `B */
	surface.w = DISP_X;
	surface.h = DISP_Y;
	surface.vbuff = vbuff;
	render_init_2d(&render, &surface, 0, 0, 0, 0);

#if defined(TRACE_ON) && (TRACE_ON)
	/* WG[o͂J܂B */
	stderr_open(1/*WG[o̓j^ڑ҂()*/);
#endif /*TRACE_ON*/

	/* NbNEEFCgؑւ܂B */
	turbo(TURBO);

	/* ؑցE16K\܂B */
	ldirect_init(LDIRECT_INTERVAL);

	/* N[Abv֐o^B */
	atexit(app_exit);

	/* eNX`܂B */
	{
		typedef const unsigned char TEXTURE_DATA[];
		extern TEXTURE_DATA TEXTURE_LIST;
		texture_table_init(texture_table, sizeof texture_table / sizeof *texture_table, TEXTURE_LIST);
		render.context->texture = &texture_table[0]; /* eNX`Œ */
	}

/*{{AvP[VL̏*/
	//srand(pceTimerGetCount());
/*}}AvP[VL̏*/
}

/****************************************************************************
 *	app_exit
 ****************************************************************************/

void
app_exit()
{
	/* ؑցE16K\܂B */
	ldirect_free();
}

/****************************************************************************
 *	schedule
 ****************************************************************************/

void
schedule()
{
#ifdef QUICK_REPEAT
	static unsigned char rep[8]; /* I[gs[gJE^ */
#endif /*QUICK_REPEAT*/

	int x;
	int y;

	pceLCDTrans();		/* ʍXV */
	yield();		/* VXe */
	joy = pcePadGet();	/* pbhԍXV */

#ifdef QUICK_REPEAT
	/* I[gs[gB */
	for(x = 0; x < 8; x++) {
		y = 1 << x; /* PAD_* */
		if(QUICK_REPEAT_PADS & y) {
			if(joy & y) {
				if(rep[x] != 0xff) rep[x]++;
				if(rep[x] >= QUICK_REPEAT_DELAY) joy |= y << 8; /* TRG_* */
			} else {
				rep[x] = 0;
			}
		}
	}
#endif QUICK_REPEAT

	now++;	/* O[o^C}XV */
}

/****************************************************************************
 *	err_disp
 ****************************************************************************/

void
err_disp(const char* fmt, ...)
{
	int len, x, y;
	char str[100]; /* ꂮ炢Ώ[H */
	va_list ap;

	va_start(ap, fmt);
	len = vsprintf(str, fmt, ap);
	va_end(ap);

	x = (DISP_X - 5 * len) / 2;
	y = (DISP_Y - 10) / 2;
	memset(vbuff, 0, sizeof vbuff);
	render_string(&render, x, y, str, 0x80, 15);
	for(;;) {
		schedule();
		if(joy & (TRG_A | TRG_B | TRG_START | TRG_SELECT)) break;
	}
}

/****************************************************************************
 *	file_init
 ****************************************************************************/

static const int bios_crc32[4] = { /* Size  =CRC-32=  =====================Filename===================== */
	0,			   /* 8192  ????????  coleco.col                                         */
	0x3aa93ef3,		   /* 8192  3aa93ef3  ColecoVision BIOS (1982).col                       */
	0x66cda476,		   /* 8192  66cda476  ColecoVision BIOS (1982) [h1] (no title delay).col */
	0x39bb16fc,		   /* 8192  39bb16fc  ColecoVision BIOS (1982) [h2] (different font).col */
};

static int file_item_compare(const void* _a, const void* _b) {
	FILEITEM* a = (FILEITEM*)_a;
	FILEITEM* b = (FILEITEM*)_b;
	return strcmp(a->fname, b->fname);
}

static void file_sort() {
	qsort(file_list, file_num, sizeof(FILEITEM), file_item_compare);
}

void
file_init()
{
	int i;
	int level;
	int len;
	FILE* fp;
	FILEINFO fi;
	FILEITEM* item;

	/* vbuff[256x192]̂128x88ȍ~̓t@CIɂ͎gȂ̂ŁA
	 * t@CXg̈Ƃėp܂B
	 * G~[Vsƃt@CXg̈悪j󂳂̂ŁA
	 * t@CXgč쐬Kv܂B
	 */
	if((128 * 88) + (sizeof(FILEITEM) * MAX_FILE) > sizeof vbuff) {
		DIE();
	}
	file_list = (FILEITEM*)(vbuff + 128 * 88);
	memset(file_list, 0, sizeof(FILEITEM) * MAX_FILE);

	/* BIOS܂B(OvACRC[0],[1],[2]vA̗D揇) */
	memset(bios_fname, 0, sizeof bios_fname);
	level = 4; /* ox: 0=OvA1..3=CRCvA4=o */
	pceFileFindOpen(&fi);
	while((strlen(bios_fname) == 0) && pceFileFindNext(&fi)) {
		if(fi.length != 0x2000) continue;
		for(i = 0; i < level; i++) {
			if(i == 0) {
				if(strcasecmp(fi.filename, "coleco.col") != 0) continue;
			} else {
				fp = fopen(fi.filename, "rb");
				if(!fp) DIE();
				fread(z80mem, 1, 0x2000, fp);
				fclose(fp);
				if(zip_crc32(z80mem, 0x2000) != bios_crc32[i]) continue;
			}
			strcpy(bios_fname, fi.filename);
			level = i;
		}
	}
	pceFileFindClose(&fi);

	/* ROM܂B */
	file_num = 0;
	item = file_list;
	pceFileFindOpen(&fi);
	while(file_num < MAX_FILE) {
		if(!pceFileFindNext(&fi)) {
			break;
		}
		if(strcmp(fi.filename, bios_fname) == 0) {
			continue;
		}
		len = strlen(fi.filename);
		if((len < 1 + 3) || (len > MAXFILENAME)) {
			continue;
		}
		if(strcasecmp(&fi.filename[len - (1 + 3)], ".col") != 0) {
			continue;
		}
		strcpy(item->fname, fi.filename);
		item->size = fi.length;
		file_num++;
		item++;
	}
	pceFileFindClose(&fi);

	/* t@CXgt@C̏ɕёւ܂B
	 * X^bNeʂƂȂ̂ŁAʃobt@g܂B
	 * 128x88=11KB΂܂vł傤B
	 */
	stack_switch(vbuff + 128 * 88, file_sort, 0, 0);
}

/****************************************************************************
 *	app_main
 ****************************************************************************/

void
app_main()
{
	int retval;
	int c;
	int i;
	int y;
	int i_file;
	FILEITEM* item;

RESET:
	/* t@CXg쐬B */
	file_init();
	if(strlen(bios_fname) == 0) {
		err_disp("BIOS܂");
		exit(0);
	}
	if(!file_num) {
		err_disp("ROM(*.col)܂");
		exit(0);
	}

	for(;;) {
		schedule();

		/* IH */
		if(joy & TRG_SELECT) {
			exit(0);
		}

		/* t@CIύXH */
		if(joy & TRG_UP) cursel--;
		if(joy & TRG_DN) cursel++;
		if(joy & TRG_LF) cursel -= MAX_DISP;
		if(joy & TRG_RI) cursel += MAX_DISP;
		if(cursel < 0                      ) cursel = 0;
		if(cursel > file_num - 1           ) cursel = file_num - 1;
		if(cursel < curtop                 ) curtop = cursel                 ;
		if(cursel > curtop + (MAX_DISP - 1)) curtop = cursel - (MAX_DISP - 1);

		/* G~[VJnH */
		if(joy & TRG_AB) {
			item = &file_list[cursel];
			mode = 1; /* G~[Vs */
			retval = coleco_run(item->fname, joy & TRG_A ? 1/*SoundOn*/ : 0/*SoundOff*/);
			mode = 0; /* t@CI */
			if(retval == 0) {
				goto RESET; /* t@CXgč쐬K{!! */
			} else {
				err_disp("sł܂ł");
			}
		}

		/* t@CXg\B */
		surface_clear(&surface, 0);
		item = &file_list[curtop];
		y = 8;
		for(i = 0; i < MAX_DISP + 1/*ɏ镪`*/; i++) {
			i_file = curtop + i;
			if(i_file > file_num - 1) break;
			if(i_file == cursel) {
				render_rectangle_fill(&render, 0, y - 1, DISP_X, 6 + 1, 10);
				c = 0;
			} else {
				c = 10;
			}
			render_string(&render, 4 + 4 *  1, y, item->fname, 0x82, c);
			render_printf(&render, 4 + 4 * 15, y, 0x82, c, "%8d bytes", item->size);
			item++;
			y += 6;
		}

		/* ^Cg\Biɂ͂ݏot@CXgɏ㏑j */
		render_rectangle_fill(&render, 0, 0, DISP_X, 6 + 1, 15);
		render_string(&render, -2, 1,
		//       012345678901234567890123456789012
			"======== ColecoVision/EMU ========",
			0x82, 0);
	}
}

/****************************************************************************
 *	test
 ****************************************************************************/

//__SIZEOF_FRAM_ADDR            $$lk33$$              uscn     global   00000c68	(C)
// N=100	117	(C)

#ifdef TEST
void
test()
{
	int i;
	int t0;
	int t1;

	for(i = 0; i < 10; i++) {
		schedule();
	}

	for(i = 0; i < sizeof coleco_dcsg; i++) {
		((unsigned char*)&coleco_dcsg)[i] = (unsigned char)(i ^ 0x77);
	}

	tms9918_reset(&coleco_vdp);
	//
	tms9918_write(&coleco_vdp, 1, 0x02);
	tms9918_write(&coleco_vdp, 1, 0x80);
	tms9918_write(&coleco_vdp, 1, 0x42);
	tms9918_write(&coleco_vdp, 1, 0x81);
	tms9918_write(&coleco_vdp, 1, 0x00);
	tms9918_write(&coleco_vdp, 1, 0x40);
	for(i = 0; i < 0x4000; i++) {
		tms9918_write(&coleco_vdp, 0, rand());
	}

DISABLE;
	t0 = pceTimerGetCount();
	for(i = 0; i < 100; i++) {
		tms9918_update(&coleco_vdp, vbuff);
	}
	t1 = pceTimerGetCount();
ENABLE;

	for(i = 0; i < sizeof coleco_dcsg; i++) {
		if(((unsigned char*)&coleco_dcsg)[i] != (unsigned char)(i ^ 0x77)) {
			DIE();
		}
	}

	tms9918_reduce(vbuff, vbuff);
	render_printf_framed(&render, 64, 44, 0x80, 0xff0, "%d", t1 - t0);
	for(;;) {
		schedule();
	}
}
#endif /*TEST*/
